/*
stattest - tests for randomness
Copyright (C) 2003  Roland Siegert, Andreas Stricker

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
*/

/** \file
 * \brief Main part and user interface for statistic test software
 *
 * This utility test a random source to its randomness.
 * The central test is the "universal statistical test"
 * Some important function like bias, entropy and histogram also
 * included
 *
 * \author Andreas Stricker
 * \author Roland Siegert
 * \version $Id: stattest.c,v 1.12 2003/05/16 14:07:33 stricand Exp $
 */

// standart includes
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define _GNU_SOURCE
#include <getopt.h>

#include "stattest.h"
#include "fileops.h"
#include "bias.h"
#include "histogram.h"
#include "entropie.h"
#include "universal.h"

// code

/** \brief Parse command line arguments.
 * Parsed command line arguments stored in structure settings.
 * \param argc Count of arguments from main
 * \param argv Vector of arguments from main
 * \param settings Pointer to settings structure where the results are stored.
 */
void parse_arguments(int argc, char *argv[],
					struct settings_struct *settings) {
	int c;
	int option_index = 0;
	bool isall_output = true;
	char last_char;

	// list of options
	static struct option long_options[] = {
		{"entropy",    0, 0, 'e'},
		{"bias",       0, 0, 'b'},
		{"histogram",  0, 0, 'i'},
		{"universal",  0, 0, 'u'},
		{"rejectrate", 1, 0, 'r'},
		{"blocksize",  1, 0, 'L'},
		{"fixedq",     0, 0, 'Q'},
		{"maxsize",    1, 0, 'm'},
		{"help",       0, 0, 'h'},
		{"version",    0, 0, 'V'},
		{0, 0, 0, 0}
	};

	// initialize with default values
	settings->is_entropy = false;
	settings->is_bias = false;
	settings->is_histogram = false;
	settings->is_universal = false;
	settings->rejection_rate = DEFAULT_REJECTION_RATE;
	settings->fixed_Q = false;
	settings->maxsize = DEFAULT_MAX_SIZE;
	settings->blocksize_L = DEFAULT_BLOCKSIZE_L;

	// start parsing
	while (1) {
//		int this_option_optind = optind ? optind : 1;
		c = getopt_long(argc, argv, "ebiur:L:Qm:hV", long_options, &option_index);
		if (c == -1) break;

		switch (c) {
		case 0:
			printf("-null-\n");
			break;
		case 'e' :
			isall_output = false;
			settings->is_entropy = true;
			break;
		case 'b' :
			isall_output = false;
			settings->is_bias = true;
			break;
		case 'i' :
			isall_output = false;
			settings->is_histogram = true;
			break;
		case 'u' :
			isall_output = false;
			settings->is_universal = true;
			break;
		case 'r' :
			if (!optarg) {
				printf(MSG_START);
				printf(MSG_ERR_ARG_REJECTRATE_NO_NUM);
				printf(MSG_HELP, argv[0]);
				exit(-1);
			}
			settings->rejection_rate = strtod(optarg, NULL);
			if ((settings->rejection_rate <= 0.0) ||
			    (settings->rejection_rate >= 0.5)) {
				printf(MSG_START);
				printf(MSG_ERR_ARG_REJECTRATE_OUT_OF_RANGE);
				printf(MSG_HELP, argv[0]);
			}
			break;
		case 'L' :
			if (!optarg) {
				printf(MSG_START);
				printf(MSG_ERR_ARG_BLOCKSIZE_NO_NUM);
				printf(MSG_HELP, argv[0]);
				exit(-1);
			}
			settings->blocksize_L = strtod(optarg, NULL);
			if (!((settings->blocksize_L == 8) ||
			    (settings->blocksize_L == 16))) {
				printf(MSG_START);
				printf(MSG_ERR_ARG_BLOCKSIZE_OUT_OF_RANGE);
				printf(MSG_HELP, argv[0]);
				exit(-1);
			}
			break;
		case 'Q' :
			settings->fixed_Q = true;
			break;
		case 'm' :
			if (!optarg) {
				printf(MSG_START);
				printf(MSG_ERR_ARG_MAXSIZE_NO_NUM);
				printf(MSG_HELP, argv[0]);
				exit(-1);
			}
			settings->maxsize = strtod(optarg, NULL);
			last_char = optarg[strlen(optarg)-1];
			switch (last_char) {
			case 'm' :
			case 'M' :
				settings->maxsize *= 1024*1024;
				break;
			case 'k' :
			case 'K' :
				settings->maxsize *= 1024;
				break;
			}
			if ((settings->maxsize <= 0) ||
			    (settings->maxsize > MAX_BUF_SIZE)) {
				printf(MSG_START);
				printf(MSG_ERR_ARG_MAXSIZE_OUT_OF_RANGE, MAX_BUF_SIZE);
				printf(MSG_HELP, argv[0]);
				exit(-1);
			}
			break;
		case 'h' :
			printf(MSG_START);
			printf(MSG_HELP, argv[0]);
			exit(0);
			break;
		case 'V' :
			printf(MSG_VERSION, VERSION);
			exit(0);
			break;
		default :
			printf(MSG_START);
			printf(MSG_ERR_ARG_UNKNOWN, c);
			printf(MSG_HELP, argv[0]);
			exit(-1);
		}
	} // while

	if (isall_output) {
		settings->is_entropy = true;
		settings->is_bias = true;
		settings->is_histogram = true;
		settings->is_universal = true;
	}

	if (optind >= argc) {
		printf(MSG_START);

		printf(MSG_ERR_ARG_NOFILE);
		printf(MSG_HELP, argv[0]);
		exit(-1);
	}

	settings->filename = (char *) calloc(strlen(argv[optind])+1, sizeof(char));
	strcpy(settings->filename, argv[optind]);

} // parse_arguments()

/** Main function
 * run the choosen tests
 */
int main(int argc, char *argv[]) {
	struct settings_struct settings;
	buffer_t buffer;
	size_t bufsize;
	double bias, entropy;
	histogram_t *histogram;
	histogram_t hist;
	struct test_params_t universal_test_result;

	histogram = &hist;		// allocate histogram array on stack

	parse_arguments(argc, argv, &settings);

	// fill buffer
	bufsize = fill_buffer_from_file(settings.filename, &buffer, settings.maxsize);
	if (bufsize == 0 && buffer == NULL) {
		free(settings.filename);
		exit(-1);
	}
	printf("Read %ul bytes\n", bufsize);

	if (settings.is_histogram || settings.is_entropy)
		histogram = build_histogram(buffer, bufsize, histogram);

	if (settings.is_histogram) {
		print_histogram(*histogram);
	}

	if (settings.is_bias) {
		bias = calc_bias(buffer, bufsize);
		printf("Bias: %e\n", bias);
	}

	if (settings.is_entropy) {
		entropy = calc_entropy(bufsize, *histogram);
		printf("Entropie: %e\n", entropy);
	}

	if (settings.is_universal) {
		universal_statistical_test(buffer, bufsize, &universal_test_result,
									settings.fixed_Q, settings.blocksize_L);
		is_universal_statistical_test_accepted(&universal_test_result,
		                                       settings.rejection_rate);
		print_universal_statistical_test(&universal_test_result);
	}

	// clean up
	free(buffer);
	free(settings.filename);
	return 0;
} // main()

