/*
stattest - tests for randomness
Copyright (C) 2003  Roland Siegert, Andreas Stricker

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
*/

/** \file
 * \brief Universal statistical test T_U
 *
 * The universal statistical test, developped by Ueli M.Maurer and
 * first appeard in Journal of Cryptology, vol. 5, no. 2, 1992, pp. 89-105
 * test a sequence U^N of its randomness.
 * The test is closly related to the entropy of the sequence and therefore
 * able to show the redundance.
 * The original Pascal sample code converted to C and enhanced to a full
 * quality test.
 *
 * \author Ueli M. Maurer
 * \author Andreas Stricker
 * \author Roland Siegert
 * \version $Id: universal.c,v 1.14 2003/05/16 14:07:33 stricand Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>	// needs library m

#include "stattest.h"
#include "universal.h"
#include "stddeviation.h"

/** Datatype for table */
struct stat_values_t {
	double expected_value;
	double variance;
};

/** Table with pre-calculatet values for different block-sizes (see Maurer) */
struct stat_values_t ust_value_table[] = {
//    expected val. Variance       Blocksize
	{ 0.7326495,    0.690    }, // 1
	{ 1.5374383,    1.338    }, // 2
	{ 2.4016068,    1.901    }, // 3
	{ 3.3112247,    2.358    }, // 4
	{ 4.2534266,    2.705    }, // 5
	{ 5.2177052,    2.954    }, // 6
	{ 6.1962507,    3.125    }, // 7
	{ 7.1836656,    3.238    }, // 8
	{ 8.1764248,    3.311    }, // 9
	{ 9.1723243,    3.356    }, // 10
	{ 10.170032,    3.384    }, // 11
	{ 11.168765,    3.401    }, // 12
	{ 12.168070,    3.410    }, // 13
	{ 13.167693,    3.416    }, // 14
	{ 14.167488,    3.419    }, // 15
	{ 15.167379,    3.421    }  // 16
};

void universal_statistical_test_8(
	buffer_t buffer,
	size_t block_size,
	struct test_params_t* test_results,
	bool fixed_Q
) {
	int *tab;
	double sum, fTU, log2_An;
	int i;
	size_t n;
	int blocks_not_found_left;
	int block_size_L = 8;
	size_t number_of_steps_K;
	size_t initial_steps_Q;
	size_t block_range_V;

	block_range_V = (0x01 << block_size_L);		// V = 2^L
	initial_steps_Q = 10*block_range_V;			// Q (Maurer recomment this value)
	number_of_steps_K = 100*block_range_V;		// K (Maurer recomment this value)

//	if (test_results == NULL) return NULL;			// return null if no valid pointer
	tab = (int*) calloc(block_range_V, sizeof(int));	// table = 2^L = V

	/* initialisation */
	for (i = 0; i < block_range_V -1; i++) tab[i] = 0;
	// Maurer suggested to do following
	// for (n = 1; n < Q; n++) tab[buffer[n-1]] = n;
	// But we know if the table is filled faster.
	// Althought let user choose with option -Q
	if (fixed_Q) {
		for (n = 1; n < initial_steps_Q; n++) tab[buffer[n-1]] = n;
	}
	else {
		blocks_not_found_left = block_range_V;
		n = 1;
		while ((blocks_not_found_left > 0) && (n < initial_steps_Q)) {
			if (tab[buffer[n-1]] == 0) --blocks_not_found_left;
			tab[buffer[n-1]] = n;
			++n;
		}
	}

	test_results->parameter_Q = n;					// initialization steps
	test_results->parameter_K =
			block_size-test_results->parameter_Q;	// effective test blocks
	test_results->parameter_L = block_size_L;		// blocksize (constant)

	/* run test */
	sum = 0.0;
//	for (n = Q+1; n < Q+K; n++) {
	for (n = test_results->parameter_Q+1; n < block_size; n++) {
		log2_An = log(n - tab[buffer[n-1]]);
		sum += log2_An;			// calculate fTU
		tab[buffer[n-1]] = n;
	}
	printf("%.10f\n", log(2.0));
	fTU = (sum / test_results->parameter_K) / log(2.0);

	test_results->ftu = fTU;						// test parameter fTU

	free(tab);		// clean up

//	return test_results;
} // universal_statistical_test_8()

void universal_statistical_test_16(
	buffer_t buffer,
	size_t block_size,
	struct test_params_t* test_results,
	bool fixed_Q
) {
	int *tab;
	double sum, fTU, log2_An;
	int i;
	size_t n;
	int blocks_not_found_left;
	int block_size_L = 16;
	size_t number_of_steps_K;
	size_t initial_steps_Q;
	size_t block_range_V;
	unsigned short int *buf;

	block_range_V = (0x01 << block_size_L);		// V = 2^L
	initial_steps_Q = 10*block_range_V;			// Q (Maurer recomment this value)
	number_of_steps_K = 100/2*block_range_V;		// K (Maurer recomment this value)

	buf = (unsigned short int*) buffer;

//	if (test_results == NULL) return NULL;			// return null if no valid pointer
	tab = (int*) calloc(block_range_V, sizeof(int));	// table = 2^L = V

	/* initialisation */
	for (i = 0; i < block_range_V -1; i++) tab[i] = 0;
	// Maurer suggested to do following
	// for (n = 1; n < Q; n++) tab[buffer[n-1]] = n;
	// But we know if the table is filled faster.
	// Althought let user choose with option -Q
	if (fixed_Q) {
		for (n = 1; n < initial_steps_Q; n++) tab[buf[n-1]] = n;
	}
	else {
		blocks_not_found_left = block_range_V;
		n = 1;
		while ((blocks_not_found_left > 0) && (n < initial_steps_Q)) {
			if (tab[buf[n-1]] == 0) --blocks_not_found_left;
			tab[buf[n-1]] = n;
			++n;
		}
	}

	test_results->parameter_Q = n;					// initialization steps
	test_results->parameter_K =
			(block_size/2)-test_results->parameter_Q;	// effective test blocks
	test_results->parameter_L = block_size_L;		// blocksize (constant)

	/* run test */
	sum = 0.0;
//	for (n = Q+1; n < Q+K; n++) {
	for (n = test_results->parameter_Q+1; n < block_size/2; n++) {
		log2_An = log(n - tab[buf[n-1]]);
		sum += log2_An;			// calculate fTU
		tab[buf[n-1]] = n;
	}
	fTU = (sum / test_results->parameter_K) / log(2.0);

	test_results->ftu = fTU;						// test parameter fTU

	free(tab);		// clean up

//	return test_results;
} // universal_statistical_test_16()

/**
 * \brief The universal statistical test is performed here.
 * The function tests the whole buffer with block_size size and write
 * the results to the structure test_results.
 * \param buffer Buffer with random data of size block_size
 * \param block_size size of buffer
 * \param test_results pointer to struct of type test_params_t, filled
 *         with the result of test and also returned by this function
 * \param block_size_L Size of a block (8 or 16)
 * \return result of test pointer to the same structure as parameter test_result
 * \sa is_universal_statistical_test_accepted(),
 *     print_universal_statistical_test()
 */
struct test_params_t* universal_statistical_test(
	buffer_t buffer,
	size_t block_size,
	struct test_params_t* test_results,
	bool fixed_Q,
	int block_size_L
) {
	switch (block_size_L) {
		case 8:
			universal_statistical_test_8(buffer, block_size, test_results,fixed_Q);
			break;
		case 16:
			universal_statistical_test_16(buffer, block_size, test_results,fixed_Q);
			break;
		default:
			printf("ERROR not defined block size\n");
			exit(-1);
	}
	return test_results;
} // universal_statistical_test()

/** A helper function for factor c(L, K)
 * c(L, K) = 0.7 - 0.8 / L + (4 + 32/L) K^(-3/L) / 15
 */
inline double ust_c_L_K(int L, size_t K) {
	return 0.7 - (0.8 / L) + (4.0 + (32.0 / L)) * (pow(K, -3.0/L) / 15);
} // ust_c_L_K()

/**
 * \brief Test if the values are accepted.
 * The test result follow normal distribution within a small range, given
 * by the standard deviation that is specified by the rejection rate.
 * This function test if the previous calculated test result is between the
 * two borders t1 and t2 and return true if test is accepted or false if not.
 * \param test_result conterns the result of test calculated by
 *         universal_statistical_test
 * \param rejection_rate specifies the rejection rate. The rate should
 *         be between ]0, 0.5]
 * \return true if test is accepted 
 *         (and also set the field accepted in test_result)
 * \sa universal_statistical_test(), print_universal_statistical_test()
 */
bool is_universal_statistical_test_accepted(
	struct test_params_t *test_result,
	double rejection_rate
) {
	double t1, t2, y, sigma;
	y = inv_standard_deviation(rejection_rate);
	sigma = ust_c_L_K(test_result->parameter_L, test_result->parameter_K) *
						sqrt(ust_value_table[BLOCK_LENGTH-1].variance /
							test_result->parameter_K);
	t1 = ust_value_table[test_result->parameter_L-1].expected_value - y * sigma;
	t2 = ust_value_table[test_result->parameter_L-1].expected_value + y * sigma;
	test_result->rejection_rate = rejection_rate;
	test_result->t1 = t1;
	test_result->t2 = t2;
	test_result->accepted = (test_result->ftu >= t1) && (test_result->ftu <= t2);
	return test_result->accepted;
} // is_universal_statistial_test_accepted()

/** Print result of universal statistical test readable to screen.
 * \param test_result a pointer to a structure test_params_t with results
 *         of the test
 */
void print_universal_statistical_test(struct test_params_t *test_result) {
	printf("\t\t True Random\tThis sequence\n");
	printf("f_Tu(U^N)]:\t%.8f\t%.8f\n",
				ust_value_table[BLOCK_LENGTH-1].expected_value,
				test_result->ftu);
	printf("Var[f_Tu(U^N)]: %e\n",
				pow(ust_c_L_K(test_result->parameter_L, test_result->parameter_K), 2.0) *
					ust_value_table[BLOCK_LENGTH-1].variance /
					test_result->parameter_K);
	printf("Parameters where: L=%u, Q=%u, K=%u\n",
				test_result->parameter_L,
				test_result->parameter_Q,
				test_result->parameter_K);
	printf("Rejection rate = %.5f ", test_result->rejection_rate);
	printf("Thresholds where: t1 = %.8f and t2 = %.8f\n",
				test_result->t1,
				test_result->t2);
	printf("Universal statistical test: %s\n",
				(test_result->accepted) ? "accepted" : "rejected");
} // print_universal_statistical_test()

